"""Settings for django-measurement."""
from appconf import AppConf
from django.conf import settings

__all__ = ("settings",)


class DjangoMeasurementConf(AppConf):
    """Settings for django-measurement."""

    BIDIMENSIONAL_SEPARATOR = "/"
    """
    For measurement classes subclassing a BidimensionalMeasure, this .
    """

    class Meta:
        prefix = "measurement"



import logging
import warnings

from django.db.models import FloatField
from django.utils.translation import ugettext_lazy as _
from measurement import measures
from measurement.base import BidimensionalMeasure, MeasureBase

from . import forms
from .utils import get_measurement

logger = logging.getLogger("django_measurement")


class MeasurementField(FloatField):
    description = "Easily store, retrieve, and convert python measures."
    empty_strings_allowed = False
    MEASURE_BASES = (
        BidimensionalMeasure,
        MeasureBase,
    )
    default_error_messages = {
        "invalid_type": _(
            "'%(value)s' (%(type_given)s) value" " must be of type %(type_wanted)s."
        ),
    }

    def __init__(
        self,
        verbose_name=None,
        name=None,
        measurement=None,
        measurement_class=None,
        unit_choices=None,
        *args,
        **kwargs
    ):

        if not measurement and measurement_class is not None:
            warnings.warn(
                '"measurement_class" will be removed in version 4.0', DeprecationWarning
            )
            measurement = getattr(measures, measurement_class)

        if not measurement:
            raise TypeError(
                "MeasurementField() takes a measurement"
                " keyword argument. None given."
            )

        if not issubclass(measurement, self.MEASURE_BASES):
            raise TypeError(
                "MeasurementField() takes a measurement keyword argument."
                " It has to be a valid MeasureBase subclass."
            )

        self.measurement = measurement
        self.widget_args = {
            "measurement": measurement,
            "unit_choices": unit_choices,
        }

        super(MeasurementField, self).__init__(verbose_name, name, *args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(MeasurementField, self).deconstruct()
        kwargs["measurement"] = self.measurement
        return name, path, args, kwargs

    def get_prep_value(self, value):
        if value is None:
            return None

        elif isinstance(value, self.MEASURE_BASES):
            # sometimes we get sympy.core.numbers.Float, which the
            # database does not understand, so explicitely convert to
            # float

            return float(value.standard)

        else:
            return super(MeasurementField, self).get_prep_value(value)

    def get_default_unit(self):
        unit_choices = self.widget_args["unit_choices"]
        if unit_choices:
            return unit_choices[0][0]
        return self.measurement.STANDARD_UNIT

    def from_db_value(self, value, *args, **kwargs):
        if value is None:
            return None

        return get_measurement(
            measure=self.measurement,
            value=value,
            original_unit=self.get_default_unit(),
        )

    def value_to_string(self, obj):
        value = self.value_from_object(obj)
        if not isinstance(value, self.MEASURE_BASES):
            return value
        return "%s:%s" % (value.value, value.unit)

    def deserialize_value_from_string(self, s: str):
        parts = s.split(":", 1)
        if len(parts) != 2:
            return None
        value, unit = float(parts[0]), parts[1]
        measure = get_measurement(self.measurement, value=value, unit=unit)
        return measure

    def to_python(self, value):

        if value is None:
            return value
        elif isinstance(value, self.MEASURE_BASES):
            return value
        elif isinstance(value, str):
            parsed = self.deserialize_value_from_string(value)
            if parsed is not None:
                return parsed
        value = super(MeasurementField, self).to_python(value)

        return_unit = self.get_default_unit()

        msg = (
            'You assigned a %s instead of %s to %s.%s.%s, unit was guessed to be "%s".'
            % (
                type(value).__name__,
                str(self.measurement.__name__),
                self.model.__module__,
                self.model.__name__,
                self.name,
                return_unit,
            )
        )
        logger.warning(msg)
        return get_measurement(measure=self.measurement, value=value, unit=return_unit,)

    def formfield(self, **kwargs):
        defaults = {"form_class": forms.MeasurementField}
        defaults.update(kwargs)
        defaults.update(self.widget_args)
        return super(MeasurementField, self).formfield(**defaults)



import pytest
from django.core import serializers
from django.core.exceptions import ValidationError
from django.utils import module_loading
from measurement import measures
from measurement.measures import Distance

from django_measurement.forms import MeasurementField
from tests.custom_measure_base import DegreePerTime, Temperature, Time
from tests.forms import (
    BiDimensionalLabelTestForm,
    LabelTestForm,
    MeasurementTestForm,
    SITestForm,
)
from tests.models import MeasurementTestModel

pytestmark = [
    pytest.mark.django_db,
]


class TestMeasurementField:
    def test_storage_of_standard_measurement(self):
        measurement = measures.Weight(g=20)

        instance = MeasurementTestModel()
        instance.measurement_weight = measurement
        instance.save()

        retrieved = MeasurementTestModel.objects.get(pk=instance.pk)

        assert retrieved.measurement_weight == measurement

    def test_storage_of_temperature(self):
        measurement = measures.Temperature(c=20)

        instance = MeasurementTestModel()
        instance.measurement_temperature = measurement
        instance.save()

        retrieved = MeasurementTestModel.objects.get(pk=instance.pk)

        assert retrieved.measurement_temperature == measurement

    def test_storage_of_string_value(self):
        instance = MeasurementTestModel()
        instance.measurement_weight = "21.4"
        instance.save()

        retrieved = MeasurementTestModel.objects.get(pk=instance.pk)

        assert retrieved.measurement_weight == measures.Weight(g=21.4)

    def test_storage_of_float_value(self):
        instance = MeasurementTestModel()
        instance.measurement_weight = 21.4
        instance.save()

        retrieved = MeasurementTestModel.objects.get(pk=instance.pk)

        assert retrieved.measurement_weight == measures.Weight(g=21.4)

    def test_storage_of_bidimensional_measurement(self):
        measurement = measures.Speed(mph=20)

        instance = MeasurementTestModel()
        instance.measurement_speed = measurement
        instance.save()

        retrieved = MeasurementTestModel.objects.get(pk=instance.pk)

        assert retrieved.measurement_speed == measurement

    def test_storage_and_retrieval_of_bidimensional_measurement(self):
        original_value = measures.Speed(mph=65)

        MeasurementTestModel.objects.create(measurement_speed=original_value,)

        retrieved = MeasurementTestModel.objects.get()

        new_value = retrieved.measurement_speed

        assert new_value == original_value
        assert type(new_value) == type(original_value)
        assert new_value.unit == original_value.STANDARD_UNIT

    def test_storage_and_retrieval_of_bidimensional_measurement_choice(self):
        original_value = measures.Speed(mph=65)

        MeasurementTestModel.objects.create(measurement_speed_mph=original_value,)

        retrieved = MeasurementTestModel.objects.get()

        new_value = retrieved.measurement_speed_mph

        assert new_value == original_value
        assert type(new_value) == type(original_value)
        assert new_value.unit == original_value.unit

    def test_storage_and_retrieval_of_measurement(self):
        original_value = measures.Weight(lb=124)

        MeasurementTestModel.objects.create(measurement_weight=original_value,)

        retrieved = MeasurementTestModel.objects.get()
        new_value = retrieved.measurement_weight

        assert new_value == original_value
        assert type(new_value) == type(original_value)
        assert new_value.unit == original_value.STANDARD_UNIT

    def test_storage_and_retrieval_of_measurement_choice(self):
        original_value = measures.Distance(km=100)

        MeasurementTestModel.objects.create(measurement_distance_km=original_value,)

        retrieved = MeasurementTestModel.objects.get()
        new_value = retrieved.measurement_distance_km

        assert new_value == original_value
        assert type(new_value) == type(original_value)
        assert new_value.unit == original_value.unit


@pytest.mark.parametrize(
    "fieldname, measure_cls",
    [
        ("measurement_weight", measures.Weight),
        ("measurement_distance", measures.Distance),
        ("measurement_distance_km", measures.Distance),
        ("measurement_speed", measures.Speed),
        ("measurement_temperature", measures.Temperature),
        ("measurement_speed_mph", measures.Speed),
        ("measurement_custom_degree_per_time", DegreePerTime),
        ("measurement_custom_temperature", Temperature),
        ("measurement_custom_time", Time),
    ],
)
class TestDeconstruct:
    def test_deconstruct(self, fieldname, measure_cls):
        field = MeasurementTestModel._meta.get_field(fieldname)

        name, path, args, kwargs = field.deconstruct()

        assert name == fieldname
        assert path == "django_measurement.models.MeasurementField"
        assert args == []
        assert kwargs["blank"] == field.blank
        assert kwargs["null"] == field.null
        assert kwargs["measurement"] == field.measurement

        new_cls = module_loading.import_string(path)
        new_field = new_cls(name=name, *args, **kwargs)

        assert type(field) == type(new_field)
        assert field.deconstruct() == (name, path, args, kwargs)


@pytest.mark.parametrize(
    "fieldname, measure, expected_serialized_value",
    [
        ("measurement_weight", measures.Weight(kg=4.0), "4.0:kg"),
        ("measurement_speed", measures.Speed(mi__hr=2.0), "2.0:mi__hr"),
    ],
)
class TestSerialization:
    def test_deconstruct(self, fieldname, measure, expected_serialized_value):
        instance = MeasurementTestModel(pk=0)
        setattr(instance, fieldname, measure)
        serialized_object = serializers.serialize("python", [instance])[0]
        serialized_value = serialized_object["fields"][fieldname]

        assert isinstance(serialized_value, str)
        assert serialized_value == expected_serialized_value

        deserialized_obj = next(serializers.deserialize("python", [serialized_object]))
        deserialized_value = getattr(deserialized_obj.object, fieldname)

        assert deserialized_value == measure


class TestMeasurementFormField:
    def test_max_value(self):
        valid_form = MeasurementTestForm(
            {"measurement_distance_0": 2.0, "measurement_distance_1": "mi"}
        )
        invalid_form = MeasurementTestForm(
            {"measurement_distance_0": 4.0, "measurement_distance_1": "mi"}
        )
        assert valid_form.is_valid()
        assert not invalid_form.is_valid()

        field = MeasurementField(measures.Distance, max_value=measures.Distance(mi=1))
        field.clean([0.5, "mi"])
        with pytest.raises(ValidationError) as e:
            field.clean([2.0, "mi"])
            assert "Ensure this value is less than or equal to 1.0 mi." in str(e)

        with pytest.raises(ValueError) as e:
            MeasurementField(measures.Distance, max_value=1.0)
            assert bytes(e) == '"max_value" must be a measure, got float'

    def test_form_storage(self):
        form = MeasurementTestForm(
            {"measurement_distance_0": 2.0, "measurement_distance_1": "mi"}
        )
        assert form.is_valid()
        obj = form.save()

        assert obj.measurement_distance == measures.Distance(mi=2)

    def test_form_bidir(self):
        form = MeasurementTestForm(
            {"measurement_speed_0": 2.0, "measurement_speed_1": "mi__hr"}
        )
        assert form.is_valid()
        obj = form.save()

        assert obj.measurement_speed == measures.Speed(mi__hr=2)

    def test_min_value(self):
        field = MeasurementField(measures.Distance, min_value=measures.Distance(mi=1.0))
        field.clean([2.0, "mi"])
        with pytest.raises(ValidationError) as e:
            field.clean([0.5, "mi"])
            assert "Ensure this value is greater than or equal to 1.0 mi." in str(e)

        with pytest.raises(ValueError) as e:
            MeasurementField(measures.Distance, min_value=1.0)
            assert str(e) == '"min_value" must be a measure, got float'

    def test_float_casting(self, caplog):
        m = MeasurementTestModel(
            measurement_distance=float(2000), measurement_distance_km=2,
        )
        m.full_clean()

        assert m.measurement_distance.value == 2000
        assert m.measurement_distance.unit == Distance.STANDARD_UNIT

        assert m.measurement_distance_km.value == 2
        assert m.measurement_distance_km.unit == "km"
        assert m.measurement_distance_km == Distance(km=2)

        m.measurement_distance_km = Distance(km=1)
        m.full_clean()
        assert m.measurement_distance_km.value == 1
        assert m.measurement_distance_km.unit == "km"
        assert m.measurement_distance_km == Distance(km=1)

        record = caplog.records[0]

        assert record.levelname == "WARNING"
        assert record.message == (
            "You assigned a float instead of Distance to"
            " tests.models.MeasurementTestModel.measurement_distance,"
            ' unit was guessed to be "m".'
        )

    def test_unicode_labels(self):
        form = LabelTestForm()
        assert ("c", u"°C") in form.fields["simple"].fields[1].choices
        assert ("f", u"°F") in form.fields["simple"].fields[1].choices
        assert ("k", u"°K") in form.fields["simple"].fields[1].choices

        si_form = SITestForm()
        assert ("ms", "ms") in si_form.fields["simple"].fields[1].choices
        assert ("Ps", "Ps") in si_form.fields["simple"].fields[1].choices

        bi_dim_form = BiDimensionalLabelTestForm()
        assert ("c__ms", u"°C/ms") in bi_dim_form.fields["simple"].fields[1].choices
        assert ("c__Ps", u"°C/Ps") in bi_dim_form.fields["simple"].fields[1].choices



from itertools import product

from django import forms
from django.core.validators import MaxValueValidator, MinValueValidator
from measurement.base import BidimensionalMeasure, MeasureBase

from django_measurement import utils
from django_measurement.conf import settings


class MeasurementWidget(forms.MultiWidget):
    def __init__(
        self,
        attrs=None,
        float_widget=None,
        unit_choices_widget=None,
        unit_choices=None,
        *args,
        **kwargs
    ):

        self.unit_choices = unit_choices

        if not float_widget:
            float_widget = forms.TextInput(attrs=attrs)

        if not unit_choices_widget:
            unit_choices_widget = forms.Select(attrs=attrs, choices=unit_choices)

        widgets = (float_widget, unit_choices_widget)
        super(MeasurementWidget, self).__init__(widgets, attrs)

    def decompress(self, value):
        if value:
            choice_units = set([u for u, n in self.unit_choices])

            unit = value.STANDARD_UNIT
            if unit not in choice_units:
                unit = choice_units.pop()

            magnitude = getattr(value, unit)
            return [magnitude, unit]

        return [None, None]


class MeasurementField(forms.MultiValueField):
    def __init__(
        self,
        measurement,
        max_value=None,
        min_value=None,
        unit_choices=None,
        validators=None,
        bidimensional_separator=settings.MEASUREMENT_BIDIMENSIONAL_SEPARATOR,
        *args,
        **kwargs
    ):

        if not issubclass(measurement, (MeasureBase, BidimensionalMeasure)):
            raise ValueError("%s must be a subclass of MeasureBase" % measurement)

        self.measurement_class = measurement
        if not unit_choices:
            if issubclass(measurement, BidimensionalMeasure):
                assert isinstance(bidimensional_separator, str), (
                    "Supplied bidimensional_separator for %s must be of string/unicode type;"
                    " Instead got type %s"
                    % (measurement, str(type(bidimensional_separator)),)
                )
                unit_choices = tuple(
                    (
                        (
                            "{0}__{1}".format(primary, reference),
                            "{0}{1}{2}".format(
                                getattr(
                                    measurement.PRIMARY_DIMENSION, "LABELS", {}
                                ).get(primary, primary),
                                bidimensional_separator,
                                getattr(
                                    measurement.REFERENCE_DIMENSION, "LABELS", {}
                                ).get(reference, reference),
                            ),
                        )
                        for primary, reference in product(
                            measurement.PRIMARY_DIMENSION.get_units(),
                            measurement.REFERENCE_DIMENSION.get_units(),
                        )
                    )
                )
            else:
                unit_choices = tuple(
                    (
                        (u, getattr(measurement, "LABELS", {}).get(u, u))
                        for u in measurement.get_units()
                    )
                )

        if validators is None:
            validators = []

        if min_value is not None:
            if not isinstance(min_value, MeasureBase):
                msg = '"min_value" must be a measure, got %s' % type(min_value)
                raise ValueError(msg)
            validators += [MinValueValidator(min_value)]

        if max_value is not None:
            if not isinstance(max_value, MeasureBase):
                msg = '"max_value" must be a measure, got %s' % type(max_value)
                raise ValueError(msg)
            validators += [MaxValueValidator(max_value)]

        float_field = forms.FloatField(*args, **kwargs)
        choice_field = forms.ChoiceField(choices=unit_choices)
        defaults = {
            "widget": MeasurementWidget(
                float_widget=float_field.widget,
                unit_choices_widget=choice_field.widget,
                unit_choices=unit_choices,
            ),
        }
        defaults.update(kwargs)
        fields = (float_field, choice_field)
        super(MeasurementField, self).__init__(
            fields, validators=validators, *args, **defaults
        )

    def compress(self, data_list):
        if not data_list:
            return None

        value, unit = data_list
        if value in self.empty_values:
            return None

        return utils.get_measurement(self.measurement_class, value, unit)



DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}}

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django_measurement",
    "tests",
)

SITE_ID = 1
ROOT_URLCONF = "core.urls"

SECRET_KEY = "foobar"

USE_L10N = True
